/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
package org.ow2.weblab.core.helper.impl;

import java.io.StringWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.extended.util.PoKUtil;
import org.ow2.weblab.core.model.PieceOfKnowledge;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;

public class AdvancedLinkedHashMap implements IPredicateValuePairs {


	private LinkedHashMap<String, List<WTriple>> map = new LinkedHashMap<String, List<WTriple>>();


	private final Statements wtm;


	private String temporaryURI = null;


	protected AdvancedLinkedHashMap(final Statements wtm, final String uri) {
		this.wtm = wtm;
		this.temporaryURI = uri;
	}


	public AdvancedLinkedHashMap(final Statements wtm) {
		this.wtm = wtm;
	}


	public LinkedHashMap<String, List<WTriple>> getMap() {
		return this.map;
	}


	public void setMap(final LinkedHashMap<String, List<WTriple>> map) {
		this.map = map;
	}


	@Override
	public Object getValue(final String predicate) {
		final List<WTriple> wts = this.map.get(predicate);
		if ((wts == null) || wts.isEmpty()) {
			return null;
		}
		if (wts.size() == 1) {
			return wts.get(0).getObject();
		}
		final List<Object> lo = new LinkedList<Object>();
		for (final WTriple wt : wts) {
			lo.add(wt.getObject());
		}
		return lo;
	}


	@Override
	public void writeStatements(final String[] predicates, final String[] objects, PieceOfKnowledge pokover) throws WebLabCheckedException {
		if ((predicates == null) || (objects == null) || (predicates.length == 0) || (objects.length == 0) || (predicates.length != objects.length)) {
			throw new WebLabCheckedException("Unable to write predicate and object.");
		}

		Collection<WTriple> wls = this.getStatements();
		if (wls.isEmpty()) {
			if (this.temporaryURI != null) {
				wls = new LinkedList<WTriple>();
				// Create a dummy WTriple
				wls.add(new WTriple(this.temporaryURI, "none", "none", null, null, pokover));
				this.temporaryURI = null;
			} else {
				throw new WebLabCheckedException("Unable to write predicate and object because there is no previous element.");
			}
		}

		final WTriple wt = wls.iterator().next();

		if (pokover == null) {
			pokover = wt.getAnnotation();
		}

		final List<WTriple> list = new LinkedList<WTriple>();

		final String uri = wt.getSubject();
		for (int i = 0; i < predicates.length; i++) {
			final WTriple nw = new WTriple(uri, predicates[i], objects[i], wt.getSubjectResource(), this.wtm.getResource(objects[i]), wt.getAnnotatedOn());
			nw.setAnnotation(pokover);
			this.put(nw);
			list.add(nw);
		}

		AdvancedLinkedHashMap.write(list, pokover, this.wtm, new HashMap<String, String>());

	}


	protected static void write(final List<WTriple> list, final PieceOfKnowledge pokover, final Statements wtm, final HashMap<String, String> mapping)
			throws WebLabCheckedException {

		final Object data = pokover.getData();

		Model model;
		if ((data == null) || !(data instanceof org.w3c.dom.Node)) {
			model = ModelFactory.createDefaultModel();
		} else {
			model = TripleSelectors.loadModel((org.w3c.dom.Node) data);
		}

		for (final WTriple wt : list) {
			boolean res = false;
			final String uri = wt.getSubject();
			final String predicate = wt.getPredicate();
			final String object = wt.getObject();
			if ((uri == null) || (predicate == null) || (object == null)) {
				throw new WebLabCheckedException("Unable to write predicate and object : " + predicate + " : " + object + " on " + uri);
			}
			for (final String namespace : wtm.namespaces) {
				if (object.startsWith(namespace)) {
					res = true;
					break;
				}
			}

			final Triple t = wt.getTriple();
			if (t != null) {
				final Node n = t.getObject();
				if ((n != null) && !n.isLiteral()) {
					res = true;
				}
			}

			String muri = mapping.get(uri);
			String mobject = mapping.get(object);
			muri = muri == null ? uri : muri;
			mobject = mobject == null ? object : mobject;

			final Resource s = model.createResource(muri);
			final Property p = model.createProperty(predicate);

			if (res) {
				model.createStatement(s, p, mobject);
			} else {

				final Statement statement = model.createStatement(s, p, t.getObject().getLiteral().getValue().toString());
				model.add(statement);
			}
		}

		final StringWriter sw = new StringWriter();
		model.write(sw);
		PoKUtil.setPoKData(pokover, sw.toString());
		model.close();
	}


	@Override
	public void writeStatement(final String predicate, final String object) throws WebLabCheckedException {
		this.writeStatements(new String[] { predicate }, new String[] { object }, null);
	}


	/**
	 * Inner function to add a WTriple
	 * 
	 * @param wt
	 *            a WTriple
	 */
	protected void put(final WTriple wt) {
		List<WTriple> list = this.map.get(wt.getPredicate());
		if (list == null) {
			list = new LinkedList<WTriple>();
		}
		list.add(wt);
		this.map.put(wt.getPredicate(), list);
	}


	@Override
	public Set<String> allPredicates() {
		return this.map.keySet();
	}


	@Override
	public Collection<WTriple> getStatements() {
		final List<WTriple> list = new LinkedList<WTriple>();
		for (final List<WTriple> li : this.map.values()) {
			list.addAll(li);
		}
		return list;

	}


	@Override
	public String toString() {
		return this.map.toString();
	}
}
