/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.core.helper.impl;

import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.core.helper.impl.Rule.Domain;
import org.ow2.weblab.core.model.PieceOfKnowledge;
import org.ow2.weblab.core.model.Resource;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.SimpleSelector;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

public class ComplexTripleSelector extends TripleSelectors implements RDFRulesSelector, AdvancedSelector, org.ow2.weblab.core.helper.impl.SimpleSelector {


	protected transient List<Rule> rules = new LinkedList<Rule>();


	private static final Log log = LogFactory.getLog(ComplexTripleSelector.class);


	public ComplexTripleSelector() {
	}


	public ComplexTripleSelector(final boolean supportNullData, final String... namespaces) {
		super(supportNullData, namespaces);
	}


	/**
	 * Add a filter on an association subject/predicat
	 * 
	 * @param subject
	 * @param predicat
	 */
	@Override
	public void addSPRule(final String subject, final String predicat) {
		// filter.addSPFilter(subject, predicat);
		this.rules.add(new Rule(Domain.SUBJECT, subject, Domain.PREDICATE, predicat));
	}


	/**
	 * Add a filter on an association subject/object
	 * 
	 * @param subject
	 * @param object
	 */
	@Override
	public void addSORule(final String subject, final String object) {
		// filter.addSPFilter(subject, object);
		this.rules.add(new Rule(Domain.SUBJECT, subject, Domain.OBJECT, object));
	}


	/**
	 * Add a filter on an association predicat/object
	 * 
	 * @param predicat
	 * @param object
	 */
	@Override
	public void addPORule(final String predicat, final String object) {
		// filter.addSPFilter(predicat, object);
		this.rules.add(new Rule(Domain.PREDICATE, predicat, Domain.OBJECT, object));
	}


	/**
	 * Find all WTriple in a Resource
	 * 
	 * @param resource the Resource
	 * @return a list of WTriple describing RDF annotations in the resource
	 */
	@Override
	public List<WTriple> findIn(final Resource resource) {
		final List<WTriple> list = super.findInResource(resource, null, null, null);
		final LinkedHashSet<WTriple> triples = new LinkedHashSet<WTriple>();
		for (final WTriple wt : list) {
			triples.add(wt);
		}
		list.clear();
		list.addAll(triples);
		return list;
	}


	/**
	 * Faster
	 */
	@Override
	protected List<WTriple> applySelection(final org.w3c.dom.Node data, final Node s, final Node p, final Node o, final String text, final Resource annotated,
			final List<Resource> resources, final PieceOfKnowledge annotation) {
		if (this.rules.isEmpty()) {
			return super.applySelection(data, s, p, o, text, annotated, resources, annotation);
		}
		final List<WTriple> list = new LinkedList<WTriple>();


		final Model model = TripleSelectors.loadModel(data);
		final Graph graph = model.getGraph();

		final StmtIterator iter = model.listStatements(new SimpleSelector(null, null, (RDFNode) null) {


			@Override
			public boolean selects(final Statement stat) {
				final Triple t = stat.asTriple();
				for (final Rule rule : ComplexTripleSelector.this.rules) {
					if (ComplexTripleSelector.this.matches(rule, t)) {
						return true;
					}
				}
				return false;
			}
		});

		while (iter.hasNext()) {
			final Object ob = iter.next();
			if (ob instanceof Statement) {
				final Triple t = ((Statement) ob).asTriple();
				final Node n = t.getSubject();
				final WTriple w = new WTriple(t, text, annotated, this.copyAndAdd(resources, annotation));
				w.setAnnotation(annotation);
				if (n.isBlank()) {
					final ExtendedIterator<?> in = graph.find(null, null, n);
					while (in.hasNext()) {
						final Object oz = in.next();
						if (oz instanceof Triple) {
							w.setReifiedWTriple(new WTriple((Triple) oz, text, annotated, resources));
						}
					}
				}
				list.add(w);
			}
		}
		model.close();
		return list;
	}


	/**
	 * Uses rules
	 */
	@Override
	public Statements searchFor(final Resource resource, final String... predicates) {
		this.rules.clear();
		final Statements res = new Statements(this.uris);
		ComplexTripleSelector.log.debug("Resource : search on " + resource);
		if (predicates == null) {
			final List<WTriple> l = this.findInResource(resource, null, null, null, true);
			res.addAll(l);
			return res;
		}

		for (final String pred : predicates) {
			this.addPORule(pred, "(.|\\s)*");
		}
		res.addAll(this.findIn(resource));
		this.rules.clear();
		return res;
	}


	@Override
	public List<WTriple> findInResource(final Resource resource, final Filter filter) {
		this.rules.clear();
		return super.findInResource(resource, filter);
	}


	@Override
	public List<WTriple> findInResource(final Resource resource, final String subject, final String predicate, final String object) {
		this.rules.clear();
		return super.findInResource(resource, subject, predicate, object);
	}


	@Override
	public List<WTriple> findInResource(final Resource resource, final String subject, final String predicate, final String object, final boolean reif) {
		this.rules.clear();
		return super.findInResource(resource, subject, predicate, object, reif);
	}


	public void addRule(final Rule rule) {
		this.rules.add(rule);
	}


	protected boolean matches(final Rule rule, final Triple t) {
		final String subject = t.getSubject().toString();
		final String predicate = t.getPredicate().toString();
		final String object = t.getObject().toString();

		String first = null;
		String second = null;
		switch (rule.first) {
			case SUBJECT:
				first = subject;
				break;
			case PREDICATE:
				first = predicate;
				break;
			case OBJECT:
				first = object;
				break;
		}
		if (rule.second != null) {
			switch (rule.second) {
				case SUBJECT:
					second = subject;
					break;
				case PREDICATE:
					second = predicate;
					break;
				case OBJECT:
					second = object;
					break;
			}
		}
		if (((first != null) && rule.firstS.equals("(.|\\s)*")) || first.matches("(?i)" + rule.firstS)) {
			if (second == null) {
				return true;
			}
			if (rule.secondS.equals("(.|\\s)*") || second.matches("(?i)" + rule.secondS)) {
				return true;
			}
		}

		return false;
	}


	@Override
	public Results select(final Resource resource, final String... predicates) {
		this.rules.clear();
		final Statements statements = this.searchFor(resource, predicates);
		return new Results(statements);
	}
}
