/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.core.helper.impl;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.helper.impl.parser.Parsers;
import org.ow2.weblab.core.helper.impl.parser.TypeParser;
import org.ow2.weblab.core.model.PieceOfKnowledge;
import org.ow2.weblab.core.model.Resource;

/**
 * An efficient structure to gather and access triples and their
 * subject/predicates/values
 * 
 * This structure supports a list of values if multiple values
 * are found for a given subject and predicate.
 * 
 * @author Arnaud
 * 
 */
public class Statements extends LinkedHashMap<String, IPredicateValuePairs> {


	/**
	 * An ascending comparator the value "a" will be after the value "b"
	 */
	public static final Comparator<String> ASC = Statements.ascComparator();


	/**
	 * A descending comparator the value "a" will be before the value "b"
	 */
	public static final Comparator<String> DESC = Statements.descComparator();


	protected Map<String, PieceOfKnowledge> pokMap = new HashMap<String, PieceOfKnowledge>();


	protected Map<String, Resource> resourcesMap = new HashMap<String, Resource>();


	protected Map<String, Resource> annotatedOn = new HashMap<String, Resource>();


	protected List<String> namespaces = new LinkedList<String>();


	public Statements(final String... namespaces) {
		for (final String namespace : namespaces) {
			this.namespaces.add(namespace);
		}
	}


	/**
	 * Returns the pok on which this subject is annotated
	 * 
	 * @param subject
	 *            a subject
	 * @return the {@link PieceOfKnowledge} in which this subject is annotated
	 */
	public PieceOfKnowledge getPokFor(final String subject) {
		return this.pokMap.get(subject);
	}


	/**
	 * Add a list of WTriple in ths Statements
	 * 
	 * @param wtl
	 *            a list of Wtriple
	 */
	public void addAll(final List<WTriple> wtl) {
		for (final WTriple wt : wtl) {
			this.add(wt);
		}
	}


	/**
	 * Return the WebLab Resource associated with the given URI.
	 * If it was found during the search.
	 * 
	 * @param uri
	 *            an URI
	 * @return a WebLab Resource
	 */
	public Resource getResource(final String uri) {
		return this.resourcesMap.get(uri);
	}


	/**
	 * Returns a resource on which a statement with this uri has been found
	 * 
	 * @param uri
	 * @return a WebLab Resource annotated with a statement concerning this uri (subject or object)
	 */
	public Resource getAnnotatedResourceWith(final String uri) {
		return this.annotatedOn.get(uri);
	}


	protected void add(final WTriple wt) {
		final WTriple wrei = wt.getReifiedWTriple();
		final String resubject = wrei == null ? null : "" + wrei.getSubject();
		final String subject = resubject == null ? wt.getSubject() : resubject;
		// String predicate = wt.getPredicate();
		final String object = wt.getObject();

		final Resource rsubject = wrei != null ? wrei.getSubjectResource() : wt.getSubjectResource();
		final Resource robject = wt.getObjectResource();

		if (wt.getAnnotation() != null) {
			this.pokMap.put(subject, wt.getAnnotation());
		}

		if (rsubject != null) {
			this.resourcesMap.put(subject, rsubject);
		}
		if (robject != null) {
			this.resourcesMap.put(object, robject);
		}

		IPredicateValuePairs value = this.get(subject);
		// Object objectvalue = object;s
		if (value == null) {
			value = new AdvancedLinkedHashMap(this);
		}
		// else{
		// // Multiple values support for the same subject and predicate
		// Object o = value.getValue(pred);
		// List<String> l = null;
		// if (o != null){
		// l = new LinkedList<String>();
		// if (o instanceof List){
		// l = (List<String>)o;
		// }else{
		// l.add(o.toString());
		// }
		// l.add(object);
		// ov = l;
		// if(l.size() == 1){
		// ov = l.iterator().next();
		// }
		// }
		// }
		final Resource res = this.annotatedOn.get(subject);
		if (res == null) {
			this.annotatedOn.put(subject, wt.getAnnotatedOn());
		}
		((AdvancedLinkedHashMap) value).put(wt);
		this.put(subject, value);
	}


	@Override
	public String toString() {
		final StringBuffer res = new StringBuffer();
		for (final String uri : this.keySet()) {
			res.append(uri + " : \n");
			final IPredicateValuePairs values = this.get(uri);
			for (final String pred : values.allPredicates()) {
				res.append(pred + " = " + values.getValue(pred) + "\n");
			}
			res.append("\n");
		}
		return res.toString();
	}


	/**
	 * Returns a list of URI in the order of predicates sorted by a comparator.
	 * If the value is not defined for a subject/predicate, its uri will be at the end if <code>defined</code> is false, else it will not be added in the final
	 * list.
	 * 
	 * @param predicate
	 *            the predicate on which value will be sorted
	 * @param comparable
	 *            the sorter
	 * @param defined
	 *            if false uri without predicate defined will be added at the end of the list, else it will be ignored
	 * @return a sorted list of URIs
	 */
	public LinkedList<String> sortBy(final String predicate, final Comparator<String> comparable, final boolean defined) {
		final LinkedList<String> listURI = new LinkedList<String>();
		final LinkedList<String> nullValues = new LinkedList<String>();
		final TreeMap<String, List<String>> tree = new TreeMap<String, List<String>>(comparable);
		for (final String uri : super.keySet()) {
			final Object value = this.get(uri).getValue(predicate);
			if (value != null) {
				List<String> suri = tree.get(value);
				if (suri == null) {
					suri = new LinkedList<String>();
				}
				suri.add(uri);
				tree.put(value.toString(), suri);
			} else if (defined) {
				nullValues.add(uri);
			}

		}

		for (final List<String> l : tree.values()) {
			listURI.addAll(l);
		}
		listURI.addAll(nullValues);
		return listURI;
	}


	/**
	 * Returns a list of URI in the order of predicates sorted by a comparator.
	 * If the value is not defined for a suject/predicate, its uri will be at the end.
	 * 
	 * @param predicate
	 *            the predicate on which value will be sorted
	 * @param comparable
	 *            the sorter
	 * @return a sorted list of URIs
	 */
	public LinkedList<String> sortBy(final String predicate, final Comparator<String> comparable) {
		return this.sortBy(predicate, comparable, false);
	}


	/**
	 * A descending comparator the value "a" will be before the value "b"
	 * 
	 * @return a descending comparator
	 */
	private static Comparator<String> descComparator() {
		return new Comparator<String>() {


			@Override
			public int compare(final String o1, final String o2) {
				return o1.compareTo(o2);
			}

		};

	}


	/**
	 * An ascending comparator the value "a" will be after the value "b"
	 * 
	 * @return an ascending comparator
	 */
	private static Comparator<String> ascComparator() {
		return new Comparator<String>() {


			@Override
			public int compare(final String o1, final String o2) {
				return o2.compareTo(o1);
			}

		};
	}


	/**
	 * Returns the first value corresponding to the predicate given.
	 * 
	 * @param subject
	 *            optional, use null as wildcard
	 * @param predicate
	 *            the uri of predicate
	 * @param locale
	 *            the locale, optional set to null as wildcard
	 * @return a value
	 */
	public String getFirstValue(final String subject, final String predicate, final Locale locale) {
		if (this.isEmpty()) {
			return null;
		}
		Object value = null;
		if (subject == null) {
			for (final String uri : this.keySet()) {
				value = this.get(uri).getValue(predicate);
				if (value != null) {

					break;
				}
			}
		} else {
			value = this.get(subject).getValue(predicate);
		}
		String result = null;
		if (value instanceof List<?>) {
			final List<String> labelList = (List<String>) value;
			for (final String label : labelList) {
				if (locale == null) {
					return this.clean(label);
				}
				if (label.contains("@" + locale.getLanguage()) || label.contains("@" + locale.getLanguage().toUpperCase(locale))) {
					result = this.clean(label);
				}
			}
		} else if (value != null) {
			if (locale == null) {
				return this.clean(value.toString());
			}
			if (((String) value).toUpperCase(locale).contains("@" + locale.getLanguage())
					|| ((String) value).toUpperCase().contains("@" + locale.getLanguage().toUpperCase(locale))) {
				result = this.clean((String) value);
			}
		}
		return result;
	}


	/**
	 * Tries to follow the RFC 4647
	 * TODO: do i better!
	 * 
	 * @param langedValue
	 * @return
	 */
	private String clean(String langedValue) {
		if ((langedValue != null) && langedValue.startsWith("\"") && langedValue.matches(".*\"@[a-zA-Z0-8-]+")) {
			final int fin = langedValue.lastIndexOf("\"");
			langedValue = langedValue.substring(1, fin);
		}
		return langedValue;
	}


	/**
	 * @deprecated Please avoid to use this thing
	 */
	@Deprecated
	public LinkedHashMap<String, LinkedHashMap<String, Object>> getMap() {
		final LinkedHashMap<String, LinkedHashMap<String, Object>> res = new LinkedHashMap<String, LinkedHashMap<String, Object>>();
		for (final String uri : this.keySet()) {
			final LinkedHashMap<String, List<WTriple>> preds = ((AdvancedLinkedHashMap) this.get(uri)).getMap();
			final LinkedHashMap<String, Object> preObj = new LinkedHashMap<String, Object>();
			for (final String pred : preds.keySet()) {
				final List<WTriple> lw = preds.get(pred);
				if (lw != null) {
					if (lw.size() == 1) {
						preObj.put(pred, lw.get(0).getObject());
					} else {
						final List<Object> lo = new LinkedList<Object>();
						for (final WTriple w : lw) {
							lo.add(w.getObject());
						}
						preObj.put(pred, lo);
					}
				}
			}
			res.put(uri, preObj);
		}
		return res;
	}


	/**
	 * Return a typed value rather than an object.
	 * 
	 * @param <T>
	 *            the type to be returned
	 * @param subject
	 *            a subject
	 * @param property
	 *            a property
	 * @param clazz
	 *            a class of the type of the value
	 * @return An instance of <T> if the value can be mapped on this type, else null.
	 */
	public <T> T getTypedValue(String subject, final String property, final Class<T> clazz) {
		if (subject == null) {
			final Set<String> set = this.getSubjects(property, ".*");
			if ((set != null) && !set.isEmpty()) {
				subject = set.iterator().next();
			}
		}

		final IPredicateValuePairs properties = this.get(subject);
		if ((properties == null) || (clazz == null)) {
			return null;
		}
		final Object value = properties.getValue(property);
		if (value == null) {
			return null;
		}
		T result = null;
		if (List.class.isAssignableFrom(value.getClass())) {
			final List l = this.getTypedValues(subject, property, clazz);
			if (l.isEmpty()) {
				return null;
			} else {
				return (T) l.get(0);
			}
		}
		result = this.parse(clazz, value, subject, property);

		return result;
	}


	private <T> T parse(final Class<T> clazz, final Object value, final String subject, final String property) throws ClassCastException {
		T result = null;
		try {
			final TypeParser<T> tp = Parsers.getParser(clazz);
			result = tp.parseValue(value);
			if ((result != null) && !clazz.isInstance(result)) {
				LogFactory.getLog(Statements.class).error(
						new ClassCastException("the value of (" + subject + ", " + property + "):" + result + " of " + result.getClass()
								+ " is not an instance of the asked class " + clazz + " !"));
				result = null;
			}
		} catch (final ParserConfigurationException e) {
			LogFactory.getLog(Statements.class).error(e);
		}
		return result;
	}


	/**
	 * Returns a list of typed values.
	 * 
	 * @param <T>
	 *            the type of the value
	 * @param subject
	 *            the uri of the subject
	 * @param property
	 *            the uri of the subject
	 * @param clazz
	 *            the class of the value
	 * @return a list of typed values.
	 */
	public <T> LinkedList<T> getTypedValues(final String subject, final String property, final Class<T> clazz) {
		final LinkedList<T> result = new LinkedList<T>();

		if (subject == null) {
			// We search for all subject

			for (final String isubject : this.keySet()) {
				result.addAll(this.getTypedValues(isubject, property, clazz));
			}


		} else {

			final IPredicateValuePairs properties = this.get(subject);
			if ((properties == null) || (clazz == null)) {
				return result;
			}
			final Object value = properties.getValue(property);
			if (value == null) {
				return result;
			}
			if (value instanceof List) {
				for (final Object o : (List) value) {
					final T parsed = this.parse(clazz, o, subject, property);
					if (parsed != null) {
						result.add(parsed);
					} else {
						LogFactory.getLog(Statements.class).warn("Object " + o + " is not defined as a subtype of " + clazz.getName());
					}
				}
			} else {
				final T parsed = this.parse(clazz, value, subject, property);
				if (parsed != null) {
					result.add(parsed);
				}
			}
		}
		return result;
	}


	/**
	 * Returns all the subjects that matches the tuple (predicate,object)
	 * 
	 * @param predicate
	 *            any URI, can be null to select all properties
	 * @param object
	 *            a regular expression or null to select all objects
	 * @return a set of subject that matches parameters
	 */
	public Set<String> getSubjects(final String predicate, String object) {

		if (object == null) {
			object = ".*";
		}
		final Set<String> res = new HashSet<String>();
		for (final String subject : this.keySet()) {
			Set<String> predicates = null;
			if (predicate == null) {
				predicates = this.get(subject).allPredicates();
			} else {
				predicates = new HashSet<String>();
				predicates.add(predicate);
			}
			for (final String pred : predicates) {
				final List<String> values = this.getTypedValues(subject, pred, String.class);
				for (final String value : values) {
					if ((value != null) && value.matches(object)) {
						res.add(subject);
						break;
					}
				}
				continue;
			}
		}
		return res;
	}


	/**
	 * Write all statements from a Statements on the given Pok and add
	 * these statements on this Statements.
	 * 
	 * @param an
	 *            other Statements
	 * @param pok
	 *            a PieceOfKnoelegde on which statements will be written
	 * @param uris
	 *            a mapping of uris. Uris are mapped using couples (previous_uri,future_uri)
	 * @throws WebLabCheckedException
	 */
	public void writeAndAdd(final Statements map, final PieceOfKnowledge pok, final String... uris) throws WebLabCheckedException {
		if ((map == null) || (pok == null)) {
			throw new WebLabCheckedException("Can not write the map :" + map + " on the PoK : " + pok);
		}
		this.resourcesMap.putAll(map.resourcesMap);
		final List<WTriple> toWrite = new LinkedList<WTriple>();

		final HashMap<String, String> mapping = new HashMap<String, String>();
		if ((uris != null) && ((uris.length % 2) == 0)) {
			for (int i = 0; i < uris.length; i += 2) {
				mapping.put(uris[i], uris[i + 1]);
			}
		}

		for (final String uri : map.keySet()) {
			final IPredicateValuePairs iplm = map.get(uri);
			String muri = mapping.get(uri);
			muri = muri == null ? uri : muri;
			IPredicateValuePairs local_iplm = this.get(muri);
			if (local_iplm == null) {
				local_iplm = new AdvancedLinkedHashMap(this, muri);
				this.put(muri, local_iplm);
			}

			final Collection<WTriple> list = iplm.getStatements();
			if (list != null) {
				toWrite.addAll(list);
			}
		}

		AdvancedLinkedHashMap.write(toWrite, pok, this, mapping);
		this.addAll(toWrite);
	}
}
