/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.core.helper.impl;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.ow2.weblab.core.extended.ontologies.RDF;
import org.ow2.weblab.core.model.Annotation;
import org.ow2.weblab.core.model.ComposedResource;
import org.ow2.weblab.core.model.Document;
import org.ow2.weblab.core.model.MediaUnit;
import org.ow2.weblab.core.model.PieceOfKnowledge;
import org.ow2.weblab.core.model.Resource;
import org.ow2.weblab.core.model.ResultSet;
import org.xml.sax.SAXParseException;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.arp.DOM2Model;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

/**
 * Selectors Helper to ease annotations retrieval.
 * Annotations triples are referenced by WTriple in which
 * subject and object can be references to WebLab Resources.
 * By default, it transforms reified triples as normal triples.
 * It is useful if you just bother on data selection.
 * Example:<br/>
 * <code>
 * 		List&lt;WTriple&gt; results = findInResource(myResource, null, "http://www.weblab-project.org/gld-country", null); 
 * </code>
 * 
 * @author EADS WebLab Team
 * @date 2009-03-11
 */
public class TripleSelectors {


	public boolean limit = false;


	public static boolean debug = false;


	public boolean supportNullData = false;


	// Multithread support
	public static boolean multi = false;


	final Object lock = new Object();


	final List<WTriple> syncList = Collections.synchronizedList(new LinkedList<WTriple>());


	private final static int numberofProc = Runtime.getRuntime().availableProcessors();


	private final static int multiplier = 1;


	int numberOfJob = 0;


	public transient String[] uris = new String[0];


	public void limitToFirstLevelAnnotation(final boolean theLimit) {
		this.limit = theLimit;
	}


	public TripleSelectors() {
		super();
	}


	public TripleSelectors(final boolean supportNullData, final String... uris) {
		this.supportNullData = supportNullData;
		if (uris == null) {
			return;
		}
		this.uris = uris;
	}


	/**
	 * Select Triplet (Subject, Predicate, Object) in the resource and
	 * its sub elements (Segments, Annotations, sub-MediaUnit ...).
	 * If subject, predicate or object is null, it acts as a wildcard parameter.
	 * Here is an example how to retrieve all Triplets which match the property "http://www.weblab-project.org/gld-country" <br/>
	 * <br/>
	 * <code>
	 * 		List&lt;WTriple&gt; results: findInResource(myResource, null, "http://www.weblab-project.org/gld-country", null);
	 * </code>
	 * 
	 * @param subject
	 *            the subject that should match in the triplet (can be null)
	 * @param predicate
	 *            the predicate that should match in the triplet (can be null)
	 * @param object
	 *            the object that should match in the triplet (can be null)
	 * @return a list of WTriple that match the subject, predicate and object.
	 */
	public List<WTriple> findInResource(final Resource resource, final String subject, final String predicate, final String object) {
		final Node s = subject == null ? Node.ANY : Node.createURI(subject);
		final Node p = predicate == null ? Node.ANY : Node.createURI(predicate);

		boolean buri = false;
		for (final String uri : this.uris) {
			if ((object != null) && object.startsWith(uri)) {
				buri = true;
				break;
			}
		}

		final Node o = object == null ? Node.ANY : buri ? Node.createURI(object) : Node.createLiteral(object);
		final List<WTriple> list = this.findInResource(resource, s, p, o, new LinkedList<Resource>());
		if (TripleSelectors.multi) {
			while (this.numberOfJob > 0) {
				Thread.yield();
				try {
					Thread.sleep(2);
				} catch (final InterruptedException e) {
					// Nothing to do
				}
			}
			list.addAll(this.syncList);
			this.syncList.clear();
		}
		return this.aggregateReified(list);
	}


	/**
	 * Select WTriple in resource according to given subject, predicate and object.
	 * Null parameters act as wildcard. It will recursively loop through Resource structure
	 * (children and their annotations) to retrieve RDF in annotations.
	 * 
	 * @param resource
	 *            a WebLab Resource on which there are annotations containing RDF
	 * @param subject
	 *            the subject that should match in the triplet (can be null)
	 * @param predicate
	 *            the predicate that should match in the triplet (can be null)
	 * @param object
	 *            the object that should match in the triplet (can be null)
	 * @param reif
	 *            if true, reified triple will not be aggregated
	 * @return a list of WTriple that match the subject, predicate and object.
	 */
	public List<WTriple> findInResource(final Resource resource, final String subject, final String predicate, final String object, final boolean reif) {
		final Node s = subject == null ? Node.ANY : Node.createURI(subject);
		final Node p = predicate == null ? Node.ANY : Node.createURI(predicate);

		boolean buri = false;
		for (final String uri : this.uris) {
			if ((object != null) && object.startsWith(uri)) {
				buri = true;
				break;
			}
		}

		final Node o = object == null ? Node.ANY : buri ? Node.createURI(object) : Node.createLiteral(object);
		if (reif) {
			final List<WTriple> list = this.findInResource(resource, s, p, o, new LinkedList<Resource>());
			if (TripleSelectors.multi) {
				list.addAll(this.syncList);
				this.syncList.clear();
			}
			return this.aggregateReified(list);
		}
		final List<WTriple> list = this.findInResource(resource, s, p, o, new LinkedList<Resource>());
		if (TripleSelectors.multi) {
			list.addAll(this.syncList);
			this.syncList.clear();
		}
		return list;
	}


	/**
	 * Select WTriple in the resource according to constrains in a filter.
	 * 
	 * @param resource
	 *            a WebLab Resource on which there are annotations containing RDF
	 * @param filter
	 *            a set of constrains
	 */
	public List<WTriple> findInResource(final Resource resource, final Filter filter) {
		final List<WTriple> list = this.findInResource(resource, null, null, null);
		if (filter == null) {
			return list;
		}
		final Set<WTriple> res = new HashSet<WTriple>();
		for (final WTriple wt : list) {
			if (!res.contains(wt) && filter.matches(wt)) {
				res.add(wt);
			}
		}
		list.clear();
		list.addAll(res);
		return list;
	}


	/**
	 * Aggregate reified triple in normal ones
	 * 
	 * @param list
	 *            a list of WTriple reified or not
	 * @return a list of WTriple not reified.
	 */
	private List<WTriple> aggregateReified(final List<WTriple> list) {
		final List<WTriple> res = new LinkedList<WTriple>(list);
		final LinkedHashMap<String, List<WTriple>> buffer = new LinkedHashMap<String, List<WTriple>>();
		for (final WTriple wt : list) {
			final String pred = wt.getPredicate();
			if (pred.startsWith("http://www.w3.org/1999/02/22-rdf-syntax-ns#")) {
				List<WTriple> inner = buffer.get(wt.getSubject());
				if (inner == null) {
					inner = new LinkedList<WTriple>();
				}
				inner.add(wt);
				buffer.put(wt.getSubject(), inner);
				// res.remove(wt);
			}
		}

		for (final Entry<String, List<WTriple>> entry : buffer.entrySet()) {
			String subject = null;
			String predicate = null;
			String object = null;
			Resource subjectr = null;
			Resource objectr = null;
			final List<WTriple> later = new LinkedList<WTriple>();
			for (final WTriple wt : entry.getValue()) {
				final String pred = wt.getPredicate();
				if (pred.equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#predicate")) {
					predicate = wt.getObject();
				} else if (pred.equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#object")) {
					object = wt.getObject();
					objectr = wt.getObjectResource();
				} else if (pred.equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#subject")) {
					subject = wt.getObject();
					subjectr = wt.getObjectResource();
				} else if (!pred.equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")) {
					later.add(wt);
				}
			}
			if ((subject != null) && (predicate != null) && (object != null)) {
				final WTriple wtriple = new WTriple(subject, predicate, object, subjectr, objectr, null);
				res.add(wtriple);
			}
			for (final WTriple wt : later) {
				res.add(new WTriple(subject, wt.getPredicate(), wt.getObject(), subjectr, wt.getObjectResource(), wt.getAnnotatedOn()));
			}
		}
		return res;
	}


	private List<WTriple> findInResource(Resource resource, final Node subject, final Node predicate, final Node object, final List<Resource> resources) {
		final List<WTriple> list = new LinkedList<WTriple>();

		if (this.limit) { // limit to the first level only
			Resource r;
			if (resource instanceof PieceOfKnowledge) {
				final PieceOfKnowledge p = new PieceOfKnowledge();
				p.setData(((PieceOfKnowledge) resource).getData());
				r = p;
			} else {
				r = new Resource();
			}
			r.setUri(resource.getUri());
			final List<Annotation> annotations = new LinkedList<Annotation>();

			// limit to first level annotation too
			for (final Annotation a : resource.getAnnotation()) {
				final Annotation c = new Annotation();
				c.setUri(a.getUri());
				c.setData(a.getData());
				annotations.add(c);
			}

			r.getAnnotation().addAll(annotations);
			resource = r;
		}

		if (resource instanceof ComposedResource) {
			list.addAll(this.find((ComposedResource) resource, subject, predicate, object, resources));

		} else if (resource instanceof MediaUnit) {
			list.addAll(this.find((MediaUnit) resource, subject, predicate, object, resources));
		} else if (resource instanceof ResultSet) {
			list.addAll(this.find((ResultSet) resource, subject, predicate, object, resources));
		} else if (resource instanceof PieceOfKnowledge) {
			list.addAll(this.find((PieceOfKnowledge) resource, subject, predicate, object, null, resource, resources));
		} else {
			for (final Annotation a : resource.getAnnotation()) {
				list.addAll(this.find(a, subject, predicate, object, null, resource, resources));
			}
		}

		return list;
	}


	/**
	 * Copies a list of resources and add an object
	 * 
	 * @param resources
	 *            list of resources
	 * @param r
	 *            a resource
	 * @return a copy the list and add the resource r
	 */
	protected List<Resource> copyAndAdd(final List<Resource> resources, final Resource r) {
		final List<Resource> copie = new LinkedList<Resource>();
		copie.addAll(resources);
		copie.add(r);
		return copie;
	}


	private List<WTriple> find(final ResultSet resultset, final Node subject, final Node predicate, final Node object, final List<Resource> resources) {
		final List<WTriple> l = new LinkedList<WTriple>();

		final PieceOfKnowledge pok = resultset.getPok();
		if (pok != null) {
			l.addAll(this.find(pok, subject, predicate, object, null, resultset, resources));
		}

		for (final Resource r : resultset.getResource()) {
			l.addAll(this.findInResource(r, subject, predicate, object, this.copyAndAdd(resources, r)));
		}

		for (final Annotation a : resultset.getAnnotation()) {
			l.addAll(this.find(a, subject, predicate, object, null, resultset, this.copyAndAdd(resources, a)));
		}
		return l;
	}


	private List<WTriple> find(final ComposedResource rc, final Node subject, final Node predicate, final Node object, final List<Resource> resources) {
		final List<WTriple> l = new LinkedList<WTriple>();

		for (final Resource r : rc.getResource()) {
			l.addAll(this.findInResource(r, subject, predicate, object, this.copyAndAdd(resources, r)));
		}

		for (final Annotation a : rc.getAnnotation()) {
			l.addAll(this.find(a, subject, predicate, object, null, rc, this.copyAndAdd(resources, a)));
		}
		return l;
	}



	private List<WTriple> find(final MediaUnit mu, final Node subject, final Node predicate, final Node object, final List<Resource> resources) {
		final List<WTriple> l = new LinkedList<WTriple>();

		resources.add(mu);

		if (mu instanceof Document) {
			final Document doc = (Document) mu;
			for (final MediaUnit mu_in_doc : doc.getMediaUnit()) {
				l.addAll(this.find(mu_in_doc, subject, predicate, object, resources));
			}
		}


		for (final Annotation a : mu.getAnnotation()) {
			l.addAll(this.find(a, subject, predicate, object, null, mu, resources));
		}

		return l;
	}



	/**
	 * Select RDF Triplet in an annotation
	 * 
	 * @param annotation
	 *            an annotation
	 * @param subject
	 *            the subject that should match in the triplet (can be null)
	 * @param predicate
	 *            the predicate that should match in the triplet (can be null)
	 * @param object
	 *            the object that should match in the triplet (can be null)
	 * @param text
	 *            the text associated with the annotation if any
	 * @param resources
	 *            the parent resources
	 * @return a list of triplet which matches the data in the annotation.
	 */
	protected List<WTriple> find(final PieceOfKnowledge annotation, final Node subject, final Node predicate, final Node object, final String text,
			final Resource annotated, final List<Resource> resources) {

		if (TripleSelectors.multi && (this.numberOfJob < (TripleSelectors.multiplier * TripleSelectors.numberofProc))) {
			return this.findMT(annotation, subject, predicate, object, text, annotated, resources);
		}

		final Object node = annotation.getData();
		if (this.supportNullData && (node == null)) {
			return new LinkedList<WTriple>();
		}

		resources.add(annotation);
		final List<WTriple> l = this.applySelection((org.w3c.dom.Node) node, subject, predicate, object, text, annotated, resources, annotation);

		for (final Annotation a : annotation.getAnnotation()) {
			l.addAll(this.find(a, subject, predicate, object, null, annotated, this.copyAndAdd(resources, a)));
		}
		return l;
	}


	protected List<WTriple> findMT(final PieceOfKnowledge annotation, final Node subject, final Node predicate, final Node object, final String text,
			final Resource annotated, final List<Resource> resources) {
		synchronized (this.lock) {
			this.numberOfJob++;
		}

		final Thread worker = new Thread("MTFind#" + this.numberOfJob) {


			@Override
			public void run() {
				final List<WTriple> list = new LinkedList<WTriple>();

				final Object node = annotation.getData();
				if (TripleSelectors.this.supportNullData && (node == null)) {
					synchronized (TripleSelectors.this.lock) {
						TripleSelectors.this.numberOfJob--;
					}
					return;
				}

				resources.add(annotation);
				list.addAll(TripleSelectors.this.applySelection((org.w3c.dom.Node) node, subject, predicate, object, text, annotated, resources, annotation));

				for (final Annotation a : annotation.getAnnotation()) {
					list.addAll(TripleSelectors.this.find(a, subject, predicate, object, null, annotated, TripleSelectors.this.copyAndAdd(resources, a)));
				}
				synchronized (TripleSelectors.this.lock) {
					TripleSelectors.this.syncList.addAll(list);
					TripleSelectors.this.numberOfJob--;
				}
			}
		};
		worker.start();

		return new LinkedList<WTriple>();
	}


	/**
	 * Load a model really fast
	 * 
	 * @param data
	 *            a DOM
	 * @return a Jena Model
	 */
	protected static Model loadModel(final org.w3c.dom.Node data) {
		final Model model = ModelFactory.createDefaultModel();
		try {
			final DOM2Model dm = DOM2Model.createD2M(RDF.NAMESPACE, model);
			dm.load(data.getFirstChild());
			dm.close();
		} catch (final SAXParseException e) {
			e.printStackTrace();
		}
		return model;
	}


	protected List<WTriple> applySelection(final org.w3c.dom.Node data, final Node s, final Node p, final Node o, final String text, final Resource annotated,
			final List<Resource> resources, final PieceOfKnowledge annotation) {
		final List<WTriple> triples = new LinkedList<WTriple>();

		final Model model = TripleSelectors.loadModel(data);
		final Graph graph = model.getGraph();

		final ExtendedIterator<?> i = graph.find(s, p, o);
		while (i.hasNext()) {
			final Object ob = i.next();
			if (ob instanceof Triple) {
				final Triple t = (Triple) ob;
				final Node n = t.getSubject();
				final WTriple w = new WTriple((Triple) ob, text, annotated, this.copyAndAdd(resources, annotation));
				w.setAnnotation(annotation);
				if (n.isBlank()) {
					final ExtendedIterator<?> in = graph.find(null, null, n);
					while (in.hasNext()) {
						final Object oz = in.next();
						if (oz instanceof Triple) {
							w.setReifiedWTriple(new WTriple((Triple) oz, text, annotated, resources));
						}
					}
				}
				triples.add(w);
			}
		}
		model.close();

		return triples;
	}
}
