/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.core.helper.impl;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;

import org.ow2.weblab.core.model.PieceOfKnowledge;
import org.ow2.weblab.core.model.Resource;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;

/**
 * A container class composed by a subject, a predicate and an object.
 * The subject and the object can be either an URI or an URI and a reference to
 * a WebLab Resource.
 * 
 * @author EADS WebLab Team
 * @date 2009-03-11
 */
public class WTriple {


	private PieceOfKnowledge annotation = null;


	private Resource annotatedOn = null;


	private WTriple reifiedWTriple = null;


	private Resource subject = null;


	private String subjectURI = null;


	private String predicate = null;


	private String objectString = null;


	private Resource object = null;


	private String annotatedText = null;


	private Triple triple;


	WTriple(final String subject, final String predicate, final String object, final Resource subjectr, final Resource objectr, final Resource annotatedOn) {
		this.annotatedOn = annotatedOn;
		this.triple = null;
		this.subjectURI = subject;
		this.subject = subjectr;
		this.predicate = predicate;
		this.objectString = this.check(object);
		this.object = objectr;

	}


	public WTriple(final Node s, final Node p, final Node o, final String annotatedText, final Resource annotatedOn, final List<Resource> resources) {
		this(new Triple(s, p, o), annotatedText, annotatedOn, resources);
	}


	public WTriple(final Triple triple, final String annotatedText, final Resource annotatedOn, final List<Resource> resources) {
		this.annotatedOn = annotatedOn;
		this.annotatedText = annotatedText;
		this.triple = triple;
		this.predicate = triple.getPredicate().toString();
		this.subjectURI = triple.getSubject().toString();
		this.objectString = this.check(triple.getObject().toString());

		boolean checked = false;
		try {
			new URI(this.objectString);
		} catch (final URISyntaxException e) {
			checked = true;
		}
		if (resources == null) {
			return;
		}
		for (final Resource r : resources) {
			if ((r != null) && (r.getUri() != null) && r.getUri().equals(this.subjectURI)) {
				this.subject = r;
				if (checked) {
					return;
				}
			}
			if (!checked && (r != null) && (r.getUri() != null) && r.getUri().equals(this.objectString)) {
				this.object = r;
				if (this.subject != null) {
					return;
				}
			}
		}
	}


	public void setAnnotation(final PieceOfKnowledge annotation) {
		this.annotation = annotation;
	}


	public PieceOfKnowledge getAnnotation() {
		return this.annotation;
	}


	/**
	 * Returns the PieceOfKnowledge in which this triplet is defined.
	 * 
	 * @return a PieceOfKnowledge
	 */
	public PieceOfKnowledge from() {
		return this.annotation;
	}


	@Override
	public int hashCode() {
		return (this.subjectURI + this.predicate + this.objectString).hashCode();
	}


	@Override
	public boolean equals(final Object o) {
		if (o instanceof WTriple) {
			final WTriple wt = (WTriple) o;
			return (this.subjectURI != null) && (this.predicate != null) && (this.objectString != null) && this.subjectURI.equals(wt.subjectURI)
					&& this.predicate.equals(wt.predicate) && this.objectString.equals(wt.objectString);
		}
		return false;

	}


	private String check(final String theObject) {
		String subobject = theObject;
		if ((theObject != null) && theObject.startsWith("\"") && theObject.endsWith("\"")) {
			subobject = theObject.substring(1, theObject.length() - 1);
		}
		return subobject;
	}


	/**
	 * Returns the Resource on which the annotation define this triple.
	 * 
	 * @return a resource
	 */
	public Resource getAnnotatedOn() {
		return this.annotatedOn;
	}


	protected void setReifiedWTriple(final WTriple reifiedWTriple) {
		this.reifiedWTriple = reifiedWTriple;
		reifiedWTriple.setAnnotation(this.annotation);
	}


	protected WTriple getReifiedWTriple() {
		return this.reifiedWTriple;
	}


	/**
	 * Return the text corresponding to the segment on which the annotation
	 * 
	 * @return the text for a segment
	 */
	public String getText() {
		return this.annotatedText;
	}


	/**
	 * Return the URI of the predicate
	 * 
	 * @return the URI
	 */
	public String getPredicate() {
		return this.predicate;
	}


	/**
	 * Return the Jena Triple use to build the WTriple
	 * 
	 * @return a Triple
	 */
	public Triple getTriple() {
		return this.triple;
	}


	/**
	 * Return the URI of the subject
	 * 
	 * @return the URI
	 */
	public String getSubject() {
		return this.subjectURI;
	}


	/**
	 * Returns the Resource if the subject is a WebLab Resource else null
	 * 
	 * @return a Resource or null
	 */
	public Resource getSubjectResource() {
		return this.subject;
	}


	/**
	 * Returns the value of the object
	 * 
	 * @return the value
	 */
	public String getObject() {
		return this.objectString;
	}


	/**
	 * Returns the Resource if the object is a WebLab Resource else null
	 * 
	 * @return a Resource or null
	 */
	public Resource getObjectResource() {
		return this.object;
	}


	@Override
	public String toString() {
		return this.subjectURI + ": " + this.subject + "\n" + this.annotatedText + " " + this.predicate + "\n" + this.objectString + ": " + this.object;
	}
}
