/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 *  
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.core.helper.impl.parser;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public final class Parsers {


	private static Map<String, TypeParser<?>> parsers = Parsers.loadParsers();


	protected static boolean strict = false;


	private static Log log = LogFactory.getLog(Parsers.class);


	private Parsers() {
		throw new UnsupportedOperationException("This class only contains " + "static methods; no need to instantiate it.");
	}


	private static HashMap<String, TypeParser<?>> loadParsers() {
		if (Parsers.log == null) {
			Parsers.log = LogFactory.getLog(Parsers.class);
		}
		Parsers.log.debug("Loading RDFType parsers.");
		final Properties prop = new Properties();
		final HashMap<String, TypeParser<?>> theParsers = new HashMap<String, TypeParser<?>>();
		try {
			InputStream is = Parsers.class.getResourceAsStream("typeParsers.properties");
			if (is == null) {
				is = Parsers.class.getResourceAsStream("conf/typeParsers.properties");
			}
			if (is == null) {
				is = ClassLoader.getSystemClassLoader().getResourceAsStream("typeParsers.properties");
			}

			if (is != null) {
				prop.load(is);
			} else {
				Parsers.log.debug("Can not load parsers from typeParsers.properties");
				Parsers.log.debug("Loading default properties:");
				Parsers.log.debug("option.strict=false\n" + "java.util.Date=" + RDFDateTypeParser.class.getCanonicalName() + "\n" + "java.lang.Integer="
						+ RDFIntegerTypeParser.class.getCanonicalName() + "\n" + "java.lang.Float=" + RDFFloatTypeParser.class.getCanonicalName() + "\n"
						+ "java.lang.String=" + RDFTransparentStringTypeParser.class.getCanonicalName());
				prop.setProperty("option.strict", "false");
				prop.setProperty("java.util.Date", RDFDateTypeParser.class.getCanonicalName());
				prop.setProperty("java.lang.Integer", RDFIntegerTypeParser.class.getCanonicalName());
				prop.setProperty("java.lang.Float", RDFFloatTypeParser.class.getCanonicalName());
				prop.setProperty("java.lang.String", RDFTransparentStringTypeParser.class.getCanonicalName());
			}
			for (final Object key : prop.keySet()) {
				Parsers.log.debug("key: " + key + " value : " + prop.getProperty(key.toString()));
				if (key.toString().startsWith("option")) {
					final Object value = prop.getProperty(key.toString());
					if ((value != null) && key.toString().endsWith("strict")) {
						Parsers.strict = value.toString().equalsIgnoreCase("true");
					}
				} else {
					theParsers.put(key.toString(), Parsers.load(prop.getProperty(key.toString())));
				}
			}
		} catch (final IOException e) {
			Parsers.log.error(e.getMessage(), e);
		}
		Parsers.log.debug("RDFType parsers loaded.");
		return theParsers;
	}


	private static TypeParser<?> load(final Object object) {
		TypeParser<?> tp = null;
		try {
			Parsers.log.debug("Try to load with System ClassLoader : " + object);
			tp = (TypeParser<?>) ClassLoader.getSystemClassLoader().loadClass(object.toString()).newInstance();
		} catch (final ClassNotFoundException e) {
			Parsers.log.debug(e.getMessage(), e);
		} catch (final InstantiationException e) {
			Parsers.log.error(e.getMessage(), e);
		} catch (final IllegalAccessException e) {
			Parsers.log.error(e.getMessage(), e);
		}
		if (tp == null) {

			try {
				Parsers.log.debug("Try to load with Local ClassLoader : " + object);
				tp = (TypeParser<?>) Parsers.class.getClassLoader().loadClass(object.toString()).newInstance();
			} catch (final ClassNotFoundException e) {
				Parsers.log.error(e.getMessage(), e);
			} catch (final InstantiationException e) {
				Parsers.log.error(e.getMessage(), e);
			} catch (final IllegalAccessException e) {
				Parsers.log.error(e.getMessage(), e);
			}
		}
		Parsers.log.debug("Loaded : " + tp);
		return tp;
	}


	public static <T> TypeParser<T> getParser(final Class<T> clazz) throws ParserConfigurationException, ClassCastException {
		final TypeParser<?> tp = Parsers.parsers.get(clazz.getName());
		if (tp == null) {
			throw new ParserConfigurationException("No parser found for type : " + clazz.getName() + " in : " + Parsers.parsers);
		}
		return (TypeParser<T>) tp;
	}
}
